// AppController -- Application controller for ThumbBorker

// Copyright (C) 2005 Borkware, LLC
// We use the BSD License.  Check out http://borkware.com/license


#import "AppController.h"
#import "BWFileList.h"
#import "BWResponderNotificationTableView.h"
#import "BWSearchArrayController.h"
#import "BWSelectableImageView.h"
#import "BWSnippetKeeper.h"
#import "BWThumbnailHTMLAccumulator.h"
#import "BWImageUtilities.h"


@implementation AppController

// some of the textfield information, such as the width and height of
// thumbnails, actually gets stored in the user defaults, so that they
// come up already filled in with their last values (expecting the
// users will be using a lot of the same settings)

- (void) setDefaultPrefs
{
    NSMutableDictionary *dictionary = [NSMutableDictionary dictionary];

    [dictionary setObject: [NSNumber numberWithInt: 64]
                forKey: @"maxWidth"];
    [dictionary setObject: [NSNumber numberWithInt: 64]
                forKey: @"maxHeight"];

    [dictionary setObject: [NSNumber numberWithInt: 1024]
                forKey: @"copyOriginalMaxWidth"];
    [dictionary setObject: [NSNumber numberWithInt: 1024]
                forKey: @"copyOriginalMaxHeight"];

    [dictionary setObject: [NSNumber numberWithBool: NO]
                forKey: @"generateHTML"];
    [dictionary setObject: [NSNumber numberWithBool: NO]
                forKey: @"copyOriginals"];

    // since the main window is using bindings to hook up the text fields,
    // set the initial values via the user defaults controller
    [[NSUserDefaultsController sharedUserDefaultsController] 
        setInitialValues: dictionary];

} // setDefaultPrefs


// given a key into the user defaults, get an int value from there

- (int) intFromDefaultsFor: (NSString *) string
{
    NSDictionary *values = [[NSUserDefaultsController 
                                sharedUserDefaultsController] values];
    NSNumber *number;
    number = [values valueForKey: string];

    return ([number intValue]);

} // intFromDefaults




// initialize a freshly allocated object

- (id) init
{
    if (self = [super init]) {
        snippets = [[BWSnippetKeeper alloc] init];
        chosenImages = [[NSMutableArray alloc] init];

        [self setDefaultPrefs];
    }

    return (self);

} // init


// clean up our mess

- (void) dealloc
{
    [snippets release];
    [fileList release];
    [chosenImages release];

    [super dealloc];
    
} // dealloc


// utility method that changes the binding of the status field and
// the image view to a new controller.  Keyboard focus is used
// to decide which tableview is the one that controls those two
// UI objects.

- (void) bindToNewController: (NSArrayController *) controller
{
    // don't unbind if we don't have to, which will clear the selection
    // in the tableview and will clear the imageView, leading to much
    // user frustration
    if (controller != lastBoundController) {

        // unhook from the previous settings
        [imageView unbind: BWSelectableImageView_valuePathBinding];
        [imageView unbind: BWSelectableImageView_selectionRectBinding];
        [imageView unbind: BWSelectableImageView_rotationFactorBinding];

        [statusField unbind: @"value"];
        [rotateCCWButton unbind: @"enabled"];
        [rotateCWButton unbind: @"enabled"];

        // fullPath is an attribute constructed on the fly in a BWFileEntry
        [statusField bind: @"value"
                     toObject: controller
                     withKeyPath: @"selection.fullPath"
                     options: nil];

        [imageView bind: BWSelectableImageView_valuePathBinding
                   toObject: controller
                   withKeyPath: @"selection.fullPath"
                   options: nil];

        // selectionRect is actually an attribute set on the BWFileEntry by
        // the imageView, and may or may not exist depending on if the
        // user has made a selection or not.

        [imageView bind: BWSelectableImageView_selectionRectBinding
                   toObject: controller
                   withKeyPath: @"selection.selectionRect"
                   options: nil];

        // rotationFactor is an NSNumber that's read by the view, but
        // tweaked by the rotate: IBAction here

        [imageView bind: BWSelectableImageView_rotationFactorBinding
                   toObject: controller
                   withKeyPath: @"selection.rotationFactor"
                   options: nil];

        // enabled setting of the rotate button depends on the existance
        // of a selected image.  Luckily the controller "canRemove" does
        // exactly this, even though we're not actually going to be 
        // removing anything

        [rotateCCWButton bind: @"enabled"
                      toObject: controller
                      withKeyPath: @"canRemove"
                      options: nil];

        [rotateCWButton bind: @"enabled"
                      toObject: controller
                      withKeyPath: @"canRemove"
                      options: nil];

        // and remember who we've just bound to, so we don't do
        // unnecessary work if asked to bind to it again
        lastBoundController = controller;
    }

} // bindToNewController


// nib file is loaded.  Do any work we can't do in Interface Builder

- (void) awakeFromNib
{
    // tell the chosenImagesController what array to manage.
    [chosenImagesController bind: @"contentArray"
                            toObject: self
                            withKeyPath: @"chosenImages"
                            options: nil];

    // the imagesController gets its contentArray bound when the user
    // picks a directory of images

    [self bindToNewController: imagesController];

    // listen to notifications from our tableview subclass when its
    // keyboard focus changes.  If it does, change which controller
    // the status line and imageView are bound to

    [[NSNotificationCenter defaultCenter]
        addObserver: self
        selector: @selector(tableViewResponderChangeNotification:)
        name: BWResponder_TableViewNotification
        object: nil];

} // awakeFromNib


// the user wants to keep an image.  Move it from the "Original Files"
// tableview down to the "Thumbnails" tableview

- (IBAction) keep: (id) sender
{
    // this is what's selected (should just be one object)
    NSArray *selectedObjects = [imagesController selectedObjects];

    // walk the list and move the object from the imagesController
    // to the chosenImagesController
    NSEnumerator *enumerator = [selectedObjects objectEnumerator];

    id thingie;
    while (thingie = [enumerator nextObject]) {
        [chosenImagesController addObject: thingie];
        [imagesController removeObject: thingie];
    }

    // by default, the new item is added to the Thumbnails list and the
    // row is selected.  For ThumbBorker, this is pretty useless behavior,
    // since once the user has evaluated an image and decided to make a
    // thumbnail, they don't really care about it anymore, and having
    // it selected is just visual distraction.  And boy is this a long
    // comment for a single line of code.
    [thumbnailsTableview deselectAll: self];

} // keep



// --------------------------------------------------
// directory reading stuff

// the user has picked a directory.  Get BWFileList to walk the
// directory and find the graphics files

- (void) readDirectory: (NSString *) fullPath
{
    // out with the old (if there is any)
    [fileList release];

    // in with the new
    fileList = [BWFileList loadFromDirectory: fullPath  deeply: YES];
    [fileList retain];

    [fileList sortByFileName];

} // readDirectory


// the open panel has closed.  If the user OK'd things, arrange to read
// the chosen directory

- (void) openPanelDidEnd: (NSOpenPanel *) panel 
	      returnCode: (int) code  
	     contextInfo: (void *) rock
{
    if (code == NSOKButton) {

        // get the filenames from the open panel
        NSArray *filenames;
	filenames = [panel filenames];

        // we're going to be creating a new contentArray eventually,
        // so forget about the current one
        [imagesController unbind: @"contentArray"];

        // walk the filenames and read them in
        NSEnumerator *enumerator = [filenames objectEnumerator];

        NSString *directory;
        while (directory = [enumerator nextObject]) {
            [self readDirectory: directory];
        }

        // now bind to the freshy loaded file entries.  This will
        // populate the tableview for us through the magic of bindings
        [imagesController bind: @"contentArray"
                          toObject: self
                          withKeyPath: @"fileList.files"
                          options: nil];
    }

} // openPanelDidEnd


// the user wants to load a directory of images.  Put up a file
// chooser and let them pick a directory where to start.

- (IBAction) loadDirectory: (id) sender
{
    NSOpenPanel *panel;
    panel = [NSOpenPanel openPanel];

    // set up the NSOpenPanel that'll just read directories
    [panel setCanChooseDirectories: YES];
    [panel setCanChooseFiles: NO];
    [panel setAllowsMultipleSelection: NO];

    // display the sheet.  We'll get called back when the user is done
    [panel beginSheetForDirectory: nil
	   file: nil
	   types: nil
	   modalForWindow: mainWindow
	   modalDelegate: self
	   didEndSelector: @selector(openPanelDidEnd:returnCode:contextInfo:)
	   contextInfo: nil];

} // loadDirectory


// --------------------------------------------------
// Thumbnail action

// original code courtesy of Michael Rothwell, from the CocoaDev wiki
// http://cocoadev.com/index.pl?ThumbnailImages

// parameters
//   source             -> the path to the image to build the thumbnail from
//   selectionRectValue -> nsvalue around an NSRect that tells us what
//                         part of the image to make a thumbnail.  Why
//			   pass in an NSValue and not an NSRect?  two reasons.
//			   The first is sloth, since the selection rectangle
//			   is obtained via a valueForKey: call, which gives
//			   us the value.  The other is because the rectangle
//			   might not exist, in which case the entire image
//			   is used.  It's cleaner to have a nil value for
//			   this than to use some illegal magic values in
//			   an NSRect
//   width, height      -> max dimensions for the thumbnail.  It's scaled
//			   proportionally to fit in these dimensions
//   quality		-> jpeg quality factor.  1.0 is the highest quality
//   saveTo             -> where to put the thumbnail image
//   scaledSize        <-  how big the scaled thumbnail is.  Can be NULL.

// (this method is a little big - should refactor into something smaller

- (BOOL) scaleAndSaveAsJPEG: (NSString *) source 
              fromRectValue: (NSValue *) selectionRectValue
                   maxWidth: (int) maxWidth 
                  maxHeight: (int) maxHeight 
             rotationFactor: (int) rotationFactor
                    quality: (float) quality
                     saveTo: (NSString *) dest
                 scaledSize: (NSSize *) scaledSize
{
    BOOL ret = NO;

    NSImage *image;
    image = BWImage_rotatedImage (source, rotationFactor);

    // could not open file
    if (!image) {
        NSLog (@"Could not load '%@'", source);
        ret = NO;
        goto bailout;
    }

    // source image size
    NSSize imageSize = [image size];

    NSRect imageBounds;
    imageBounds.origin = NSZeroPoint;
    imageBounds.size = imageSize;

    // limit the portion of the image used to the selection rectangle,
    // if there is one
    if (selectionRectValue != nil) {
        NSRect rect;
        [selectionRectValue getValue: &rect];

        // don't make thumbnails from tiny rects
        if (rect.size.width > 10 && rect.size.height > 10) {
            imageBounds = rect;
            imageSize = rect.size;
        }
    }

    int width, height;
    int newWidth, newHeight;

    width = newWidth = imageSize.width;
    height = newHeight = imageSize.height;

    if (width > maxWidth || height > maxHeight) {
        float widthRatio, heightRatio;

        // ratios
        widthRatio = width / (float)maxWidth;
        heightRatio = height / (float)maxHeight;

        if (widthRatio > heightRatio) { // landscape
            newWidth = maxWidth;
            newHeight = height / widthRatio;

        } else { // portrait
            newHeight = maxHeight;
            newWidth = width / heightRatio;
        }
    }

    if (scaledSize != NULL) {
        scaledSize->width = newWidth;
        scaledSize->height = newHeight;
    }

    // image to render into
    NSImage *scratch = nil;
    scratch = [[NSImage alloc] 
                   initWithSize: NSMakeSize(newWidth, newHeight)];
    [scratch autorelease];

    // could not create image
    if (!scratch) {
        NSLog (@"Could not render image");
        ret = NO;
        goto bailout;
    }

    // draw into image, to scale it
    [scratch lockFocus];
    [[NSGraphicsContext currentContext] 
        setImageInterpolation: NSImageInterpolationHigh];

    NSRect imageRect = NSMakeRect(0, 0, newWidth, newHeight);

    [[NSColor whiteColor] set];
    [NSBezierPath fillRect: imageRect];

    [image drawInRect: imageRect
           fromRect: imageBounds
           operation: NSCompositeSourceOver
           fraction: 1.0];

    NSBitmapImageRep *output = nil;
    output = [[NSBitmapImageRep alloc] 
                 initWithFocusedViewRect: imageRect];
    [output autorelease];
    [scratch unlockFocus];

    // could not get result
    if (!output) {
        NSLog (@"Could not scale image");
        ret = NO;
        goto bailout;
    }

    // save as JPEG
    NSDictionary *properties =
        [NSDictionary dictionaryWithObjectsAndKeys:
                          [NSNumber numberWithFloat: quality],
                      NSImageCompressionFactor, NULL];    

    NSData *bitmapData;
    bitmapData = [output representationUsingType: NSJPEGFileType
                         properties: properties];

    // could not get result
    if (bitmapData == nil) {
        NSLog (@"Could not convert to JPEG");
        ret = NO;
        goto bailout;
    }

    ret = [bitmapData writeToFile: dest  atomically: YES];

bailout:
    return (ret);

} // scaleAndSaveAsJPEG


// accessors for getting the values the user put into the textfields

- (int) maxWidth
{
    return ([self intFromDefaultsFor: @"maxWidth"]);

} // maxWidth


- (int) maxHeight
{
    return ([self intFromDefaultsFor: @"maxHeight"]);

} // maxHeight


- (int) copyOriginalMaxWidth
{
    return ([self intFromDefaultsFor: @"copyOriginalMaxWidth"]);

} // copyOriginalMaxWidth


- (int) copyOriginalMaxHeight
{
    return ([self intFromDefaultsFor: @"copyOriginalMaxHeight"]);

} // copyOriginalMaxHeight


// utility method called when the save panel is closed.
// destination is where the thumbnails are to be written.  This method
// walks the selectedObjects array and makes thumbnails for them.
// It also accumulates the HTML page (if desired by the user)

- (void) saveThumbnailsInto: (NSString *) destination
{
    BWThumbnailHTMLAccumulator *html = nil;

    if ([self intFromDefaultsFor: @"generateHTML"]) {
        // if the user wants to generate HTML, make a new accumulator.
        // if the user doens't want to generate HTML, the subsequent
        // code takes advantage of the "messages to nil are no-ops"
        // behavior of Objective-C, which simplifies the code a lot
        // not having to check everywhere if we're generating HTML or
        // not
        html = [BWThumbnailHTMLAccumulator accumulatorWithSnippets: snippets];
    }

    [html addHeader];
    [html addBeforeThumbnails];

    // walk the selected entries for the thumbnails
    NSArray *selectedObjects = [chosenImagesController arrangedObjects];
    NSEnumerator *enumerator = [selectedObjects objectEnumerator];

    id fileEntry;
    while (fileEntry = [enumerator nextObject]) {
        NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

        // where to get the data from
        NSString *path;
        path = [fileEntry valueForKey: @"fullPath"];

        // name of the file without all that leading path stuff
        NSString *filename;
        filename = [fileEntry valueForKey: @"fileName"];

        // any selection rect?  nil if not
        NSValue *selectionRectValue;
        selectionRectValue = 
            [fileEntry valueForKey: BWSelectableImageView_selectionRectBinding];

        // any rotation factor?  nil if not
        NSNumber *rotationFactorNumber;
        rotationFactorNumber =
            [fileEntry valueForKey: BWSelectableImageView_rotationFactorBinding];

        int rotationFactor = [rotationFactorNumber intValue];

        // filename without extension.  The file's going to be turned
        // into a jpeg, so we don't need to keep the extension around
        NSString *basename;
        basename = [filename stringByDeletingPathExtension];

        // construct the thumbnail file name as
        // destination path + file basename + _th + .jpg
        NSString *thumbnailFileName;
        thumbnailFileName = [NSString stringWithFormat:
                                          @"%@/%@_th.jpg",
                                      destination, basename];

        // how big the thumbnail actually is.  Need to know this to
        // generate the proper html width/height attributes to the <img>
        // tag
        NSSize scaledSize;

        // now actually scale the image and save it

        if (![self scaleAndSaveAsJPEG: path
                   fromRectValue: selectionRectValue
                   maxWidth: [self maxWidth]
                   maxHeight: [self maxHeight]
                   rotationFactor: rotationFactor
                   quality: 1.0
                   saveTo: thumbnailFileName
                   scaledSize: &scaledSize]) {
            NSLog (@"problem saving");
        }

        // if the user wanted to copy originals, do so.
        // The images get run through the scaling pipeline too,
        // since often they're insanely too huge to be useful
        // on a webpage

        if ([self intFromDefaultsFor: @"copyOriginals"]) {

            // build the destination file name
            // destination path + file basename + .jpg
            NSString *copiedOriginalFileName;
            copiedOriginalFileName = [NSString stringWithFormat:
                                                   @"%@/%@.jpg",
                                               destination, basename];

            if (![self scaleAndSaveAsJPEG: path
                       fromRectValue: nil
                       maxWidth: [self copyOriginalMaxWidth]
                       maxHeight: [self copyOriginalMaxHeight]
                       rotationFactor: rotationFactor
                       quality: 1.0
                       saveTo: copiedOriginalFileName
                       scaledSize: nil]) {
                NSLog (@"problem saving");
            }

            // this will add an <a href=> tag to the html to point
            // to the copied file
            [html addImageNamed: copiedOriginalFileName
                  thumbnailNamed: thumbnailFileName
                  size: scaledSize];

        } else {
            // just adds the thumbnail with no linking
            [html addImageNamed: nil
                  thumbnailNamed: thumbnailFileName
                  size: scaledSize];
        }
        [pool release];
    }

    [html addAfterThumbnails];
    [html addFooter];

    // now have the html accumulator actually write the file
    NSString *htmlFile;
    htmlFile = [destination stringByAppendingString: @"/index.html"];

    [html writeToFile: htmlFile];
    
} // saveThumbnailsInto


// callback from the open panel.  If the user OK's it, then
// write the goodies to the given directory

- (void) saveThumbnailsPanelDidEnd: (NSOpenPanel *) panel
                        returnCode: (int) code
                       contextInfo: (void *) rock
{
    if (code == NSOKButton) {
        NSArray *filenames;
	filenames = [panel filenames];

        // BWFileList doesn't really support multiple directories yet
        NSEnumerator *enumerator = [filenames objectEnumerator];

        NSString *directory;
        while (directory = [enumerator nextObject]) {
            [self saveThumbnailsInto: directory];
            break;
        }
    }

} // saveThumbnailsPanelDidEnd


// user wants thumbnails!  make them!  now!

- (IBAction) makeThumbnails: (id) sender
{
    NSOpenPanel *panel;

    // set up an NSOpenPanel that'll just read directories.
    // yeah, an OpenPanel.  The Save Panel prompts for a name, and
    // we don't really need that.

    panel = [NSOpenPanel openPanel];

    [panel setCanCreateDirectories: YES];
    [panel setCanChooseDirectories: YES];
    [panel setCanChooseFiles: NO];
    [panel setAllowsMultipleSelection: NO];
    [panel setPrompt: @"Save Thumbnails"];

    [panel beginSheetForDirectory: nil
	   file: nil
	   types: nil
	   modalForWindow: mainWindow
	   modalDelegate: self
	   didEndSelector: @selector(saveThumbnailsPanelDidEnd:returnCode:contextInfo:)
	   contextInfo: nil];

} // makeThumbnails   




// --------------------------------------------------
// other random housekeeping

// user wants to edit the html snippets.  Show the editing window
// in a sheet

- (IBAction) showHTMLSnippets: (id) sender
{
    [snippets showSnippetsAsSheetInWindow: mainWindow];

} // showHTMLSnippets


- (IBAction) rotate: (id) sender
{
    NSNumber *currentRotationFactor;
    currentRotationFactor = [lastBoundController 
                                valueForKeyPath: @"selection.rotationFactor"];
    int currentFactor = 0;

    if (currentRotationFactor != nil) {
        currentFactor = [currentRotationFactor intValue];
    }

    currentFactor += [sender tag];

    [lastBoundController setValue: [NSNumber numberWithInt: currentFactor]
                         forKeyPath: @"selection.rotationFactor"];
} // rotate


// our special tableviews broadcast a notification when they gain
// keyboard focus.  When a tableview gains keyboard focus, change
// the bindings of the image view and the status line so that it
// feeds from the controller that also feeds off the selection
// from that tableview

- (void) tableViewResponderChangeNotification: (NSNotification *) notification
{
    id object = [notification object];

    if (object == originalFilesTableview) {

        [self bindToNewController: imagesController];
        [thumbnailsTableview deselectAll: self];

    } else if (object == thumbnailsTableview) {

        [self bindToNewController: chosenImagesController];
        [originalFilesTableview deselectAll: self];
    }

} // tableViewResponderChangeNotification



// it's a single-window non-document app.  Nothing to do if the window
// gets closed, so quit the program when it gets closed

- (BOOL) applicationShouldTerminateAfterLastWindowClosed: (NSApplication *) sender
{
    return (YES);

} // applicationShouldTerminateAfterLastWindowClosed



// work around toolkit bug where a menu item is hooked up to an
// action, but is bound somewhere else for its "enabled" state.
// Even though that somewhere else says "disabled", if it has a
// connection to an action, it'll use the existance of that connection
// to decide on enabled/disabled state

- (BOOL) validateMenuItem: (id) menuItem
{
    BOOL result = YES;

    if ([menuItem action] == @selector(keep:)) {

        // ask the images controller if we can remove items.
        // this is the binding that was desired in the first place

        NSNumber *value;
        value = [imagesController valueForKey: @"canRemove"];
        result = [value boolValue];
    }

    return (result);

} // validateMenuItem

@end // AppController


