// BWFileList -- maintains lists of graphic files

// Copyright (C) 2005 Borkware, LLC
// We use the BSD License.  Check out http://borkware.com/license

#import "BWFileList.h"

@implementation BWFileEntry


// make a new file entry

+ (BWFileEntry *) newEntryWithPath: (NSString *) path
                           basedAt: (NSString *) at
{
    NSFileManager *manager = [NSFileManager defaultManager];
    BWFileEntry *entry = nil;

    if ([manager fileExistsAtPath: path]) {
	NSDictionary *attributes = [manager fileAttributesAtPath: path
					    traverseLink: YES];
	int size = [[attributes objectForKey: NSFileSize] intValue];

	entry = [[BWFileEntry alloc] initWithRoot: at
				     relativePath: path
				     fileName: [path lastPathComponent]
				     fileSize: size];
    }

    [entry autorelease];

    return (entry);
    
} // newEntryWithPath


// initialize a new file entry

- (id) initWithRoot: (NSString *) r  relativePath: (NSString *) rpath
	   fileName: (NSString *) name   fileSize: (int) size
{
    if (self = [super init]) {
	root = [r copy];
	relativePath = [rpath copy];
        fileName = [name copy];
	fileSize = size;
    }

    return (self);

} // initWithEverything, hold the pickles



// clean up our mess

- (void) dealloc
{
    [root release];
    [relativePath release];
    [fileName release];
    [properties release];

    [super dealloc];

} // dealloc



// this gets used with %@ in NSLog and "po" in gdb

- (NSString *) description
{
    NSString *desc;
    desc = [NSString stringWithFormat: @"name: %@  size: %d  relativePath: %@",
		     fileName, fileSize, relativePath];

    return (desc);

} // description



// return the full path to the file referenced by the file entry

- (NSString *) fullPath
{
    return ([root stringByAppendingFormat: @"/%@", relativePath]);

} // fullPath



// the file name, which may have path components

- (NSString *) fileName
{
    return (fileName);

} // fileName


// the size of the file, computed at the time it was found during the
// directory walk

- (int) fileSize
{
    return (fileSize);

} // fileSize



// used for sorting the file array

- (NSComparisonResult) compareByName: (BWFileEntry *) thing2
{
    return ([[self fileName] caseInsensitiveCompare: [thing2 fileName]]);
} // compareByName



// If someone uses KVC to set a property we don't know about, go ahead
// and hang on to it.  Lets clients add additional goodies referencing
// the file

- (void) setValue: (id) value  forUndefinedKey: (NSString *) key
{
    if (properties == nil) {
        properties = [[NSMutableDictionary alloc] init];
    }

    [properties setValue: value  forKey: key];

} // setValue forUndefinedKey



// if someone is using KVC, they can retrieve properties they set 
// previously.

- (id) valueForUndefinedKey: (NSString *) key
{
    id value;
    value = [properties valueForKey: key];

    return (value);

} // valueForUndefinedKey


@end // BWFileEntry



@implementation BWFileList

+ (BWFileList *) loadFromDirectory: (NSString *) pathName
                            deeply: (BOOL) deeply
{
    NSMutableArray *files;
    files = [NSMutableArray arrayWithCapacity: 5000];

    // access to the file system
    NSFileManager *manager = [NSFileManager defaultManager];

    // file system walker
    NSDirectoryEnumerator *dirent;
    dirent = [manager enumeratorAtPath: pathName];

    // descend in to subfolders?
    if (!deeply) {
	[dirent skipDescendents];
    }

    // figure out the file types (file extensions) we want
    NSMutableArray *imageFileTypes;
    imageFileTypes = [[NSMutableArray alloc] init];

    // these are the kinds of files desired (graphics files)
    [imageFileTypes addObjectsFromArray: [NSImage imageFileTypes]];

    // I don't want pdfs.  deal with it.
    [imageFileTypes removeObject: @"pdf"];
    [imageFileTypes removeObject: @"PDF"];

    [manager changeCurrentDirectoryPath: pathName];

    // now actually walk the directories
    NSString *file;
    while (file = [dirent nextObject]) {
	NSString *extension = [file pathExtension];

	if ([imageFileTypes indexOfObject: extension] != NSNotFound) {
	    BWFileEntry *entry;
	    entry = [BWFileEntry newEntryWithPath: file  basedAt: pathName];
	    [files addObject: entry];
	}
    }
    [imageFileTypes release];

    // now turn the array of file entires into a file list
    BWFileList *list;
    list = [[BWFileList alloc] initWithBasePath: pathName
                               relativeFileList: files];

    return (list);

} // loadFromDirectory



// init a new file list

- (id) initWithBasePath: (NSString *) base 
       relativeFileList: (NSArray *) list
{
    if (self = [super init]) {
	basePath = [base retain];
	fileList = [list retain];
    }

    return (self);

} // initWithBasePath



// how many file elements we have

- (int) count
{
    return ([fileList count]);

} // count



// reference a file

- (BWFileEntry *) getFileAt: (int) index
{
    return ([fileList objectAtIndex: index]);

} // getFileAt


// remove a file from the list

- (void) removeFileAt: (int) index
{
    [fileList removeObjectAtIndex: index];;
} // removeFileAt



// order the files by their names

- (void) sortByFileName
{
    [fileList sortUsingSelector: @selector(compareByName:)];

} // sortByFileName


// KVC stuff, providing KVC access to the "files" attribute

- (unsigned int) countOfFiles
{
    return ([fileList count]);
} // countOfFiles


- (id) objectInFilesAtIndex: (unsigned int) index
{
    return ([fileList objectAtIndex: index]);

} // objectInFilesAtIndex


- (void) insertObject: (id) object  inFilesAtIndex: (unsigned int) index
{
    // it's a nuke-only collection
} // insertObjectAtIndex


- (void) removeObjectFromFilesAtIndex: (unsigned int) index
{
    [fileList removeObjectAtIndex: index];
} // removeFromFilesAtIndex

@end // BWFileList


