// BWSelectableImageView -- An image view that lets users select regions.

// Copyright (C) 2005 Borkware, LLC
// We use the BSD License.  Check out http://borkware.com/license

#import "BWSelectableImageView.h"
#import "BWImageUtilities.h"

// names of the bindings

NSString *BWSelectableImageView_valuePathBinding = @"valuePath";
NSString *BWSelectableImageView_selectionRectBinding = @"selectionRect";
NSString *BWSelectableImageView_rotationFactorBinding = @"rotationFactor";



@implementation BWSelectableImageView

// gets called before any of our objects get created.  Expose
// the bindings we support

+ (void) initialize
{
    [self exposeBinding: BWSelectableImageView_valuePathBinding];
    [self exposeBinding: BWSelectableImageView_selectionRectBinding];

} // initialize



// init method

- (id) initWithFrame: (NSRect) frame
{
    if (self = [super initWithFrame: frame]) {
        transform = [[NSAffineTransform alloc] init];
        
        dragMode = BWSelectableImageView_NotDragging;
    }

    return (self);

} // initWithFrame



// clean up our mess

- (void) dealloc
{
    [transform release];
    [inverse release];
    [valuePathKeyPath release];
    [selectionRectKeyPath release];
    [rotationFactorKeyPath release];
    [lastImagePath release];

    [image release];
    
    [super dealloc];

} // dealloc



// these are the bindings that we support.  Probably only useful if we're
// in an IB palette

- (NSArray *) exposedBindings
{
    return ([NSArray arrayWithObjects: BWSelectableImageView_valuePathBinding,
                     BWSelectableImageView_selectionRectBinding, 
                     BWSelectableImageView_rotationFactorBinding, 
                     nil]);

} // exposedBindings



// query the binding for the path to use for the image

- (NSString *) valuePath
{
    id thing = [valuePathBinding valueForKeyPath: valuePathKeyPath];

    // if there's no selection in the binding target, just return nil
    if (thing == NSNoSelectionMarker) {
        thing = nil;
    }

    return (thing);

} // valuePath



// query the binding for the rotation factor

- (NSNumber *) rotationFactor
{
    id thing = [rotationFactorBinding valueForKeyPath: rotationFactorKeyPath];

    // if there's no selection, return zero

    if (thing == NSNoSelectionMarker) {
        thing = [NSNumber numberWithInt: 0];
    }

    return (thing);

} // rotationFactor


// this might be unnecessary
- (int) rotationFactorAsInt
{
    NSNumber *number = [self rotationFactor];
    
    return ([number intValue]);

} // rotationFactorAsInt



// take the selection anchor and opposite and build a legal NSRect with
// them

- (NSRect) normalizedSelectionRect
{
    NSRect rect;

    rect.origin.x = MIN (selectionAnchor.x, selectionOpposite.x);
    rect.origin.y = MIN (selectionAnchor.y, selectionOpposite.y);

    rect.size.width = fabsf(selectionOpposite.x - selectionAnchor.x);
    rect.size.height = fabsf(selectionOpposite.y - selectionAnchor.y);

    return (rect);

} // normalizedSelectionRect



// get the selection rectangle

- (NSRect) selectionRect
{
    return ([self normalizedSelectionRect]);

} // selectionRect



// take the selection rect and break it into the anchor and opposite
// points (forming the diagonal across the square).  Used when a
// new selection rectangle comes across the binding.

- (void) setSelectionRect: (NSRect) rect
{
    selectionAnchor.x = NSMinX (rect);
    selectionAnchor.y = NSMaxY (rect);

    selectionOpposite.x = NSMaxX (rect);
    selectionOpposite.y = NSMinY (rect);

} // setSelectionRect



// utility method that ideally should be in some common library
// or provided by Apple.
// Given two rectangles, center one inside of the other

- (NSRect) centerRect: (NSRect) smallRect
               inRect: (NSRect) bigRect
{
    NSRect centerRect;
    centerRect.size = smallRect.size;

    centerRect.origin.x = (bigRect.size.width - smallRect.size.width) / 2.0;
    centerRect.origin.y = (bigRect.size.height - smallRect.size.height) / 2.0;

    return (centerRect);

} // centerRect



// fit a rectangle inside of another one so that it maximally fill up
// one of the dimension

typedef enum RectAxis {
    RectAxisHorizontal,
    RectAxisVertical
} RectAxis;


- (NSRect) fitRect: (NSRect) inner
              into: (NSRect) outer 
         dimension: (RectAxis) dimension
{
    NSRect result;
    float proportion;

    if (dimension == RectAxisHorizontal) {
	proportion = inner.size.height / inner.size.width;
	result.size.width = outer.size.width;
	result.size.height = result.size.width * proportion;
	result.origin = inner.origin;

    } else if (dimension == RectAxisVertical) {
	proportion = inner.size.width / inner.size.height;
	result.size.height = outer.size.height;
	result.size.width = result.size.height * proportion;
	result.origin = inner.origin;

    } else {
	NSLog (@"fitRect:  bad dimension");
	result = NSZeroRect;
    }

    return (result);

} // fitRect with dimension



// center one rectangle into another one

- (NSRect) centerRect: (NSRect) inner  in:  (NSRect) outer
{
    NSRect result;

    float iwidth = inner.size.width;
    float iheight = inner.size.height;
    float owidth = outer.size.width;
    float oheight = outer.size.height;

    float x, y;

    x = outer.origin.x + (owidth - iwidth) / 2;
    y = outer.origin.y + (oheight - iheight) / 2;

    result = NSMakeRect (x, y, iwidth, iheight);
    
    return (result);

} // centerRect



// how many rectangle fitting methods can we have?  at least one
// more.

- (NSRect) fitRect: (NSRect) inner  into: (NSRect) outer
{
    NSRect result;
    float innerProportion;

    innerProportion = inner.size.width / inner.size.height;

    result = inner;
    if (innerProportion > 1) {
	// wider
	result = [self fitRect: inner
                       into: outer
                       dimension: RectAxisHorizontal];
    } else {
	// taller
	result = [self fitRect: inner
                       into: outer
                       dimension: RectAxisVertical];
    }

    if (result.size.width > outer.size.width) {
	result = [self fitRect: inner
                       into: outer
                       dimension: RectAxisHorizontal];
    }
    if (result.size.height > outer.size.height) {
	result = [self fitRect: inner
                       into: outer
                       dimension: RectAxisVertical];
    }
    if (result.size.width > outer.size.width) {
	result = [self fitRect: inner
                       into: outer
                       dimension: RectAxisHorizontal];
    }

    result = [self centerRect: result
                   in: outer];

    return (result);

} // fitRect



// we have a new image, so recalculate the transform to match it

- (void) resetTransform
{
    if (image != nil) {
        NSRect imageRect;
        imageRect.origin = NSZeroPoint;
        imageRect.size = [image size];
        
        NSRect fitRect;
        fitRect = [self fitRect: imageRect
                        into: [self bounds]];
        [transform release];
        
        transform = [[NSAffineTransform alloc] init];
        
        [transform scaleXBy: imageRect.size.width / fitRect.size.width
                   yBy: imageRect.size.height / fitRect.size.height];
        
        [transform translateXBy: -fitRect.origin.x  
                   yBy: -fitRect.origin.y];
        
        [inverse release];
        inverse = [transform copy];
        [inverse invert];
    }

} // resetTransform



// load the image at the given path

- (void) setImageWithPath: (NSString *) path
{
    NSString *valuePath = [self valuePath];

    if (valuePath != nil) {
        if (![valuePath isEqualToString: lastImagePath]
            || lastImageRotationFactor != [self rotationFactorAsInt]) {

            [image release];
            image = nil;

            image = BWImage_rotatedImage (valuePath, [self rotationFactorAsInt]);
            [image retain];

            [lastImagePath release];
            lastImagePath = [valuePath copy];
            lastImageRotationFactor = [self rotationFactorAsInt];

            // reset the selection rect as well
            selectionAnchor = selectionOpposite = NSZeroPoint;
        }
        // otherwise just eat the change
    }

    [self resetTransform];
    [self setNeedsDisplay: YES];

} // setImageWithPath



// draw the view

- (void) drawRect: (NSRect) rect
{
    NSRect bounds = [self bounds];

    // white background
    [[NSColor whiteColor] setFill];
    [NSBezierPath fillRect: bounds];

    NSRect imageRect;
    imageRect.origin = NSZeroPoint;
    imageRect.size = [image size];

    NSRect destRect = [self fitRect: imageRect
                            into: bounds];
    // draw a gray border around the picture so it's obvious where it
    // ends, in case the image has a lot of white in in
    NSColor *clearGray;
    clearGray = [NSColor colorWithDeviceWhite: 0.7 alpha: 0.5];

    if (image != nil) {
        [image drawInRect: destRect
               fromRect: imageRect
               operation: NSCompositeSourceOver
               fraction: 1.0];
        
        [clearGray set];
        [NSBezierPath strokeRect: destRect];
    }

    // the only time these should be equal is when set to be a zero
    // point, which in floating point comparison is equal

    if (selectionAnchor.x != selectionOpposite.x
        && selectionAnchor.y != selectionOpposite.y) {

        // draw the selection rectangle

        NSRect selection = [self normalizedSelectionRect];

        // go from image coordinates to view coordinates
        NSRect selectionRect;
        selectionRect.origin = [inverse transformPoint: selection.origin];
        selectionRect.size = [inverse transformSize: selection.size];

        NSBezierPath *selectionPath = [NSBezierPath bezierPathWithRect: bounds];
        [selectionPath setWindingRule: NSEvenOddWindingRule];
        [selectionPath appendBezierPathWithRect: selectionRect];
        
        [selectionPath fill];
    }

    // thin black border around the whole thing

    [[NSColor blackColor] setStroke];
    [NSBezierPath strokeRect: bounds];

} // drawRect


// bindings vector through here.  For each binding, hang on to the
// key path folks are binding us to.

- (void) bind: (NSString *) bindingName
     toObject: (id) observableObject
  withKeyPath: (NSString *) observableKeyPath
      options: (NSDictionary *) options
{
    if ([bindingName 
            isEqualToString: BWSelectableImageView_valuePathBinding]) {
        // hold on to the binding info
        valuePathBinding = observableObject;
        valuePathKeyPath = [observableKeyPath copy];

        // connect KVO
        [valuePathBinding addObserver: self
                 forKeyPath: valuePathKeyPath
                 options: nil
                 context: NULL];

        // new binding, need to redraw
        [self setNeedsDisplay: YES];
    }
    
    if ([bindingName isEqualToString:
                         BWSelectableImageView_selectionRectBinding]) {
        // hold on to the binding info
        selectionRectBinding = observableObject;
        selectionRectKeyPath = [observableKeyPath copy];
        
        // connect KVO
        [valuePathBinding addObserver: self
                          forKeyPath: selectionRectKeyPath
                          options: nil
                          context: NULL];
        
        // new binding, needs to redraw
        [self setNeedsDisplay: YES];
    }

    if ([bindingName isEqualToString:
                         BWSelectableImageView_rotationFactorBinding]) {
        // hold on to the binding info
        rotationFactorBinding = observableObject;
        rotationFactorKeyPath = [observableKeyPath copy];
        
        // connect KVO
        [valuePathBinding addObserver: self
                          forKeyPath: rotationFactorKeyPath
                          options: nil
                          context: NULL];
        
        // new binding, needs to redraw
        [self setNeedsDisplay: YES];
    }

} // bind



// detach a binding

- (void) unbind: (NSString *) bindingName
{
    if ([bindingName isEqualToString:
                         BWSelectableImageView_valuePathBinding]) {
        // disconnect the KVO
        [valuePathBinding removeObserver: self
                 forKeyPath: valuePathKeyPath];
        valuePathBinding = nil;

        // clean up the key path we've kept
        [valuePathKeyPath release];
        valuePathKeyPath = nil;

        // the current image isn't value with respect to the binding,
        // so get rid of it
        [image release];
        image = nil;
    
        [self setNeedsDisplay: YES];
    }

    if ([bindingName isEqualToString:
                         BWSelectableImageView_selectionRectBinding]) {
        // disconnect the KVO
        [selectionRectBinding removeObserver: self
                 forKeyPath: selectionRectKeyPath];
        selectionRectBinding = nil;

        // clean up the key path we've kept
        [selectionRectKeyPath release];
        selectionRectKeyPath = nil;
    
        [self setNeedsDisplay: YES];
    }

    if ([bindingName isEqualToString:
                         BWSelectableImageView_rotationFactorBinding]) {
        // disconnect the KVO
        [rotationFactorBinding removeObserver: self
                 forKeyPath: rotationFactorKeyPath];
        rotationFactorBinding = nil;

        // clean up the key path we've kept
        [rotationFactorKeyPath release];
        rotationFactorKeyPath = nil;
    
        [self setNeedsDisplay: YES];
    }

} // unbind



// KVO notification.  One of our bindees has had a change, so track it

- (void) observeValueForKeyPath: (NSString *) kp
                       ofObject: (id) object
                         change: (NSDictionary *) change
                        context: (void *) context
{
    // see what bindee thing changed

    if (object == valuePathBinding) {
        // new image.  woo!
        [self setImageWithPath: [self valuePath]];
    }

    if (object == selectionRectBinding) {
        // new selection rectangle.  woo!
        NSValue *value = [object valueForKeyPath: selectionRectKeyPath];

        // don't do anything if there's no selection
        if (value != NSNoSelectionMarker && value != nil) {
            NSRect rect;
            [value getValue: &rect];

            [self setSelectionRect: rect];
        }
    }

    if (object == rotationFactorBinding) {
        [self setImageWithPath: [self valuePath]];
        [self setNeedsDisplay: YES];
    }

} // observeValueForKeyPath


// event handling

// user clicked

- (void) mouseDown: (NSEvent *) event
{
    dragMode = BWSelectableImageView_NotDragging;

    // no need to do any work if we don't have an imagex
    if (image == nil) {
        return;
    }

    dragMode = BWSelectableImageView_Normal;

    // use the current mouse point (converted to image coordinates)
    // as the anchor
    NSPoint mouse = [self convertPoint: [event locationInWindow]
                          fromView: nil];
    mouse = [transform transformPoint: mouse];

    // first see if the mouse point is inside of the selection rectangle
    if (NSPointInRect (mouse, [self normalizedSelectionRect])) {
        dragMode = BWSelectableImageView_MoveSelection;
        
        selectionTrack = mouse;

    } else {

        // shift-click constrains selection to a square
        if ([event modifierFlags] & NSShiftKeyMask) {
            dragMode = BWSelectableImageView_Constrain;
        }
        
        selectionAnchor = selectionOpposite = mouse;
    }

    // update in case the selection changed (which it probably did)
    [self setNeedsDisplay: YES];

} // mouseDown


// called during a drag

- (void) mouseDragged: (NSEvent *) event
{
    if (image == nil) {
        return;
    }

    // get the mouse coordinate in image coordinates
    NSPoint mouse = [self convertPoint: [event locationInWindow]
                          fromView: nil];
    mouse = [transform transformPoint: mouse];

    // shift-key constrains to a square
 if (dragMode == BWSelectableImageView_MoveSelection) {
        float deltaX, deltaY;
        deltaX = mouse.x - selectionTrack.x;
        deltaY = mouse.y - selectionTrack.y;
        
        selectionAnchor.x += deltaX;
        selectionAnchor.y += deltaY;

        selectionOpposite.x += deltaX;
        selectionOpposite.y += deltaY;

        selectionTrack = mouse;
    } else {

        // this is the opposite corner from the anchor
        selectionOpposite = mouse;
            
        if (dragMode == BWSelectableImageView_Constrain) {

            // OmniGraffle uses the vertical displacement for deciding
            // what the "square size" is going to be.  Good enough for them,
            // good enough for me
            
            float height = fabsf(selectionAnchor.y - selectionOpposite.y);
            
            if (mouse.x < selectionAnchor.x) {
                selectionOpposite.x = selectionAnchor.x - height;
            } else {
                selectionOpposite.x = selectionAnchor.x + height;
            }
            
        }
    }

    [self setNeedsDisplay: YES];

} // mouseDragged



// mouse released

- (void) mouseUp: (NSEvent *) event
{
    // no need to do any work if we don't have an imagex
    if (image == nil) {
        return;
    }

    // figure out the selection rectangle
    NSRect selectionRect = [self normalizedSelectionRect];

    // wrap in a value and tell the bound object the new value

    NSValue *value;
    value = [NSValue valueWithBytes: &selectionRect
                     objCType: @encode(NSRect)];

    [selectionRectBinding setValue: value
                          forKeyPath: selectionRectKeyPath];
} // mouseUp



// keep the view <-> image coordinate transform up to date when we
// resize

- (void) setFrame: (NSRect) frame
{
    [super setFrame: frame];
    [self resetTransform];

} // setFrame


- (void) resetCursorRects
{
    [super resetCursorRects];
    [self addCursorRect: [self bounds]
          cursor: [NSCursor crosshairCursor]];

} // resetCursorRects

@end // BWSelectableImageView
